package sync

import (
	"fmt"
	"path"
	"strings"
	"unicode"

	"srs.epita.fr/fic-server/libfic"
)

// isFullGraphic detects if some rune are not graphic one.
// This function is usefull to display warning when importing key ending with \r.
func isFullGraphic(s string) bool {
	for _, c := range s {
		if !unicode.IsGraphic(c) {
			return false
		}
	}
	return true
}

// SyncExerciceKeys reads the content of flags.txt and import them as Key for the given challenge.
func SyncExerciceKeys(i Importer, exercice fic.Exercice) []string {
	var errs []string

	if _, err := exercice.WipeKeys(); err != nil {
		errs = append(errs, err.Error())
	} else if flags, err := getFileContent(i, path.Join(exercice.Path, "flags.txt")); err != nil {
		errs = append(errs, fmt.Sprintf("%q: unable to read flags: %s", path.Base(exercice.Path), err))
	} else {
		for nline, flag := range strings.Split(flags, "\n") {
			flag_splt := strings.SplitN(string(flag), "\t", 2)
			if len(flag_splt) > 2 {
				errs = append(errs, fmt.Sprintf("%q: error in flags file at line %d: invalid format", path.Base(exercice.Path), nline + 1))
				continue
			}

			if len(flag_splt) == 1 {
				errs = append(errs, fmt.Sprintf("%q: error in flags file at line %d: no separator found", path.Base(exercice.Path), nline + 1))
				continue
			}

			var label, rawkey string
			if len(flag_splt[0]) == 0 {
				label = "Flag"
			} else {
				label = flag_splt[0]
			}
			rawkey = flag_splt[1]

			if !isFullGraphic(rawkey) {
				errs = append(errs, fmt.Sprintf("%q: WARNING in flags file at line %d: non-printable characters in flag, is this really expected?", path.Base(exercice.Path), nline + 1))
			}

			if _, err := exercice.AddRawKey(label, rawkey); err != nil {
				errs = append(errs, fmt.Sprintf("%q: error in flags file at line %d: %s", path.Base(exercice.Path), nline + 1, err))
				continue
			}

		}
	}
	return errs
}

// SyncExerciceMCQ reads the content of flags-ucq.txt and flags-mcq.txt, and import them as MCQs for the given challenge.
func SyncExerciceMCQ(i Importer, exercice fic.Exercice) (errs []string) {
	if _, err := exercice.WipeMCQs(); err != nil {
		errs = append(errs, err.Error())
		return errs
	}

	// Unique Choice Questions (checkbox)
	if ucq, err := getFileContent(i, path.Join(exercice.Path, "flags-ucq.txt")); err != nil {
		if i.exists(path.Join(exercice.Path, "flags-ucq.txt")) {
			errs = append(errs, fmt.Sprintf("%q: unable to read ucq: %s", path.Base(exercice.Path), err))
		}
	} else if flag, err := exercice.AddMCQ(""); err != nil {
		errs = append(errs, fmt.Sprintf("%q: unable to add ucq: %s", path.Base(exercice.Path), err))
	} else {
		for nline, quest := range strings.Split(ucq, "\n") {
			if len(quest) == 0 {
				continue
			}

			if len(quest) < 2 {
				errs = append(errs, fmt.Sprintf("%q: error in ucq file at line %d: missing response", path.Base(exercice.Path), nline + 1))
				continue
			}

			// Expect 0 or 1
			if quest[0] != 48 && quest[0] != 49 {
				errs = append(errs, fmt.Sprintf("%q: error in ucq file at line %d: invalid format: first character has to be either 0 or 1", path.Base(exercice.Path), nline + 1))
				continue
			}

			if _, err := flag.AddEntry(quest[1:], quest[0] == 49); err != nil {
				errs = append(errs, fmt.Sprintf("%q: error in ucq file at line %d: %s", path.Base(exercice.Path), nline + 1, err))
			}
		}
	}

	// Multiple Choice Questions (radio)
	if mcq, err := getFileContent(i, path.Join(exercice.Path, "flags-mcq.txt")); err != nil {
		if i.exists(path.Join(exercice.Path, "flags-mcq.txt")) {
			errs = append(errs, fmt.Sprintf("%q: unable to read mcq: %s", path.Base(exercice.Path), err))
		}
	} else {
		for nline, quest := range strings.Split(mcq, "\n") {
			quest_splt := strings.Split(string(quest), "\t")
			if len(quest_splt) < 2 {
				errs = append(errs, fmt.Sprintf("%q: error in mcq file at line %d: not enough responses", path.Base(exercice.Path), nline + 1))
				continue
			}

			if flag, err := exercice.AddMCQ(quest_splt[0]); err != nil {
				errs = append(errs, fmt.Sprintf("%q: error in mcq file at line %d: %s", path.Base(exercice.Path), nline + 1, err))
				continue
			} else {
				hasOne := false
				for cid, choice := range quest_splt[1:] {
					// Expect 0 or 1
					if choice[0] != 48 && choice[0] != 49 {
						errs = append(errs, fmt.Sprintf("%q: error in mcq file at line %d,%d: invalid format: first character has to be either 0 or 1", path.Base(exercice.Path), nline + 1, cid))
						continue
					}

					if len(quest) < 2 {
						errs = append(errs, fmt.Sprintf("%q: error in mcq file at line %d: missing label", path.Base(exercice.Path), nline + 1))
						continue
					}

					if _, err := flag.AddEntry(choice[1:], choice[0] == 49); err != nil {
						errs = append(errs, fmt.Sprintf("%q: error in mcq file at line %d,%d: %s", path.Base(exercice.Path), nline + 1, cid, err))
						continue
					}

					if choice[0] == 49 {
						hasOne = true
					}
				}
				if !hasOne {
					errs = append(errs, fmt.Sprintf("%q: warning in mcq file at line %d: no valid answer defined, is this really expected?", path.Base(exercice.Path), nline + 1))
				}
			}
		}
	}

	return errs
}
